@@ -108,13 +108,24 @@ static void at_registration_status(struct ofono_netreg *netreg,
 
 	cbd->user = nd;
 
-	/*
-	 * Send *ERINFO to find out the current tech, it will be intercepted
-	 * in mbm_erinfo_notify
-	 */
-	if (nd->vendor == OFONO_VENDOR_MBM)
+	switch (nd->vendor) {
+	case OFONO_VENDOR_MBM:
+		/*
+		 * Send *ERINFO to find out the current tech, it will be
+		 * intercepted in mbm_erinfo_notify
+		 */
 		g_at_chat_send(nd->chat, "AT*ERINFO?", none_prefix,
 				NULL, NULL, NULL);
+		break;
+	case OFONO_VENDOR_NOVATEL:
+		/*
+		 * Send $CNTI=0 to find out the current tech, it will be
+		 * intercepted in nw_cnti_notify
+		 */
+		g_at_chat_send(nd->chat, "AT$CNTI=0", none_prefix,
+				NULL, NULL, NULL);
+		break;
+	}
 
 	if (g_at_chat_send(nd->chat, "AT+CREG?", creg_prefix,
 				at_creg_cb, cbd, g_free) > 0)
@@ -140,10 +151,8 @@ static void cops_cb(gboolean ok, GAtResult *result, gpointer user_data)
 
 	decode_at_error(&error, g_at_result_final_response(result));
 
-	if (!ok || nd->mcc[0] == '\0' || nd->mnc[0] == '\0') {
-		cb(&error, NULL, cbd->data);
-		goto out;
-	}
+	if (!ok)
+		goto error;
 
 	g_at_result_iter_init(&iter, result);
 
@@ -180,14 +189,12 @@ static void cops_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	DBG("cops_cb: %s, %s %s %d", name, nd->mcc, nd->mnc, tech);
 
 	cb(&error, &op, cbd->data);
-
-out:
 	g_free(cbd);
 
 	return;
 
 error:
-	CALLBACK_WITH_FAILURE(cb, NULL, cbd->data);
+	cb(&error, NULL, cbd->data);
 
 	g_free(cbd);
 }
@@ -196,9 +203,14 @@ static void cops_numeric_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct cb_data *cbd = user_data;
 	struct netreg_data *nd = ofono_netreg_get_data(cbd->user);
+	ofono_netreg_operator_cb_t cb = cbd->cb;
 	GAtResultIter iter;
 	const char *str;
 	int format;
+	int len;
+	struct ofono_error error;
+
+	decode_at_error(&error, g_at_result_final_response(result));
 
 	if (!ok)
 		goto error;
@@ -215,19 +227,31 @@ static void cops_numeric_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	if (ok == FALSE || format != 2)
 		goto error;
 
-	if (g_at_result_iter_next_string(&iter, &str) == FALSE ||
-		strlen(str) == 0)
+	if (g_at_result_iter_next_string(&iter, &str) == FALSE)
+		goto error;
+
+	len = strspn(str, "0123456789");
+
+	if (len != 5 && len != 6)
 		goto error;
 
 	extract_mcc_mnc(str, nd->mcc, nd->mnc);
 
 	DBG("Cops numeric got mcc: %s, mnc: %s", nd->mcc, nd->mnc);
 
-	return;
+	ok = g_at_chat_send(nd->chat, "AT+COPS=3,0", none_prefix,
+					NULL, NULL, NULL);
+
+	if (ok)
+		ok = g_at_chat_send(nd->chat, "AT+COPS?", cops_prefix,
+					cops_cb, cbd, NULL);
+
+	if (ok)
+		return;
 
 error:
-	nd->mcc[0] = '\0';
-	nd->mnc[0] = '\0';
+	cb(&error, NULL, cbd->data);
+	g_free(cbd);
 }
 
 static void at_current_operator(struct ofono_netreg *netreg,
@@ -249,14 +273,6 @@ static void at_current_operator(struct ofono_netreg *netreg,
 		ok = g_at_chat_send(nd->chat, "AT+COPS?", cops_prefix,
 					cops_numeric_cb, cbd, NULL);
 
-	if (ok)
-		ok = g_at_chat_send(nd->chat, "AT+COPS=3,0", none_prefix,
-					NULL, NULL, NULL);
-
-	if (ok)
-		ok = g_at_chat_send(nd->chat, "AT+COPS?", cops_prefix,
-					cops_cb, cbd, NULL);
-
 	if (ok)
 		return;
 
@@ -517,14 +533,14 @@ static void option_osigq_notify(GAtResult *result, gpointer user_data)
 				at_util_convert_signal_strength(strength));
 }
 
-static void option_owcti_notify(GAtResult *result, gpointer user_data)
+static void option_ouwcti_notify(GAtResult *result, gpointer user_data)
 {
 	int mode;
 	GAtResultIter iter;
 
 	g_at_result_iter_init(&iter, result);
 
-	if (!g_at_result_iter_next(&iter, "_OWCTI:"))
+	if (!g_at_result_iter_next(&iter, "_OUWCTI:"))
 		return;
 
 	if (!g_at_result_iter_next_number(&iter, &mode))
@@ -624,6 +640,24 @@ static void option_ossysi_notify(GAtResult *result, gpointer user_data)
 	ofono_info("OSSYSI mode: %d", mode);
 }
 
+static void huawei_rssi_notify(GAtResult *result, gpointer user_data)
+{
+	struct ofono_netreg *netreg = user_data;
+	GAtResultIter iter;
+	int strength;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^RSSI:"))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &strength))
+		return;
+
+	ofono_netreg_strength_notify(netreg,
+				at_util_convert_signal_strength(strength));
+}
+
 static void csq_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct cb_data *cbd = user_data;
@@ -744,6 +778,31 @@ static void mbm_erinfo_notify(GAtResult *result, gpointer user_data)
 	}
 }
 
+static void nw_cnti_notify(GAtResult *result, gpointer user_data)
+{
+	//struct ofono_netreg *netreg = user_data;
+	//struct netreg_data *nd = ofono_netreg_get_data(netreg);
+	GAtResultIter iter;
+	const char *tech;
+	int option;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (g_at_result_iter_next(&iter, "$CNTI:") == FALSE)
+		return;
+
+	if (g_at_result_iter_next_number(&iter, &option) == FALSE)
+		return;
+
+	if (option != 0)
+		return;
+
+	if (g_at_result_iter_next_unquoted_string(&iter, &tech) == FALSE)
+		return;
+
+	ofono_info("CNTI: %s", tech);
+}
+
 static void creg_notify(GAtResult *result, gpointer user_data)
 {
 	struct ofono_netreg *netreg = user_data;
@@ -850,13 +909,15 @@ static void at_creg_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	case OFONO_VENDOR_OPTION_HSO:
 		g_at_chat_send(nd->chat, "AT_OSSYS=1", none_prefix,
 				NULL, NULL, NULL);
+		g_at_chat_send(nd->chat, "AT_OUWCTI=1", none_prefix,
+				NULL, NULL, NULL);
 		g_at_chat_send(nd->chat, "AT_OCTI=1", none_prefix,
 				NULL, NULL, NULL);
 		g_at_chat_send(nd->chat, "AT_OSQI=1", none_prefix,
 				NULL, NULL, NULL);
 		g_at_chat_register(nd->chat, "_OSIGQ:", option_osigq_notify,
 					FALSE, netreg, NULL);
-		g_at_chat_register(nd->chat, "_OWCTI:", option_owcti_notify,
+		g_at_chat_register(nd->chat, "_OUWCTI:", option_ouwcti_notify,
 					FALSE, netreg, NULL);
 		g_at_chat_register(nd->chat, "_OCTI:", option_octi_notify,
 					FALSE, netreg, NULL);
@@ -865,28 +926,13 @@ static void at_creg_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 
 		g_at_chat_send(nd->chat, "AT_OSSYS?", none_prefix,
 				NULL, NULL, NULL);
+		g_at_chat_send(nd->chat, "AT_OWCTI?", none_prefix,
+				NULL, NULL, NULL);
 		g_at_chat_send(nd->chat, "AT_OCTI?", none_prefix,
 				NULL, NULL, NULL);
 		g_at_chat_send(nd->chat, "AT_OSQI?", none_prefix,
 				NULL, NULL, NULL);
-
-		/*
-		 * Option has the concept of Speech Service versus
-		 * Data Service. Problem is that in Data Service mode
-		 * the card will reject all voice calls. This is a
-		 * problem for Multi-SIM cards where one of the SIM
-		 * cards is used in a mobile phone and thus incoming
-		 * calls would be not signalled on the phone.
-		 *
-		 *   0 = Speech Service enabled
-		 *   1 = Data Service only mode
-		 */
-		g_at_chat_send(nd->chat, "AT_ODO?", none_prefix,
-				NULL, NULL, NULL);
-		g_at_chat_send(nd->chat, "AT_ODO=0", none_prefix,
-				NULL, NULL, NULL);
 		break;
-
 	case OFONO_VENDOR_MBM:
 		g_at_chat_send(nd->chat, "AT*ERINFO=1", none_prefix,
 				NULL, NULL, NULL);
@@ -895,11 +941,18 @@ static void at_creg_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		g_at_chat_send(nd->chat, "AT+CIND=?", cind_prefix,
 				cind_support_cb, netreg, NULL);
 		return;
-	case OFONO_VENDOR_HUAWEI:
+	case OFONO_VENDOR_NOVATEL:
 		/*
-		 * Huawei doesn't support CIND, signal strength reported
-		 * in the modem driver
+		 * Novatel doesn't support unsolicited notifications
+		 * of technology changes, but register a handle for
+		 * CNTI so we get notified by any query.
 		 */
+		g_at_chat_register(nd->chat, "$CNTI:", nw_cnti_notify,
+					FALSE, netreg, NULL);
+		break;
+	case OFONO_VENDOR_HUAWEI:
+		g_at_chat_register(nd->chat, "^RSSI:", huawei_rssi_notify,
+					FALSE, netreg, NULL);
 		break;
 	default:
 		g_at_chat_send(nd->chat, "AT+CIND=?", cind_prefix,
@@ -990,8 +1043,8 @@ static struct ofono_netreg_driver driver = {
 	.name				= "atmodem",
 	.probe				= at_netreg_probe,
 	.remove				= at_netreg_remove,
-	.registration_status 		= at_registration_status,
-	.current_operator 		= at_current_operator,
+	.registration_status		= at_registration_status,
+	.current_operator		= at_current_operator,
 	.list_operators			= at_list_operators,
 	.register_auto			= at_register_auto,
 	.register_manual		= at_register_manual,
