@@ -39,8 +39,6 @@
 
 #define uninitialized_var(x) x = x
 
-#define SMS_MANAGER_INTERFACE "org.ofono.SmsManager"
-
 #define SMS_MANAGER_FLAG_CACHED 0x1
 
 #define SETTINGS_STORE "sms"
@@ -60,7 +58,6 @@ struct ofono_sms {
 	unsigned int next_msg_id;
 	guint ref;
 	GQueue *txq;
-	time_t last_mms;
 	gint tx_source;
 	struct ofono_message_waiting *mw;
 	unsigned int mw_watch;
@@ -70,6 +67,7 @@ struct ofono_sms {
 	const struct ofono_sms_driver *driver;
 	void *driver_data;
 	struct ofono_atom *atom;
+	ofono_bool_t use_delivery_reports;
 };
 
 struct pending_pdu {
@@ -105,7 +103,7 @@ static void set_sca(struct ofono_sms *sms,
 	value = phone_number_to_string(&sms->sca);
 
 	ofono_dbus_signal_property_changed(conn, path,
-						SMS_MANAGER_INTERFACE,
+						OFONO_SMS_MANAGER_INTERFACE,
 						"ServiceCenterAddress",
 						DBUS_TYPE_STRING, &value);
 }
@@ -134,6 +132,9 @@ static DBusMessage *generate_get_properties_reply(struct ofono_sms *sms,
 	ofono_dbus_dict_append(&dict, "ServiceCenterAddress", DBUS_TYPE_STRING,
 				&sca);
 
+	ofono_dbus_dict_append(&dict, "UseDeliveryReports", DBUS_TYPE_BOOLEAN,
+				&sms->use_delivery_reports);
+
 	dbus_message_iter_close_container(&iter, &dict);
 
 	return reply;
@@ -206,7 +207,7 @@ static void sca_set_callback(const struct ofono_error *error, void *data)
 	struct ofono_sms *sms = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Setting SCA failed");
+		DBG("Setting SCA failed");
 		__ofono_dbus_pending_reply(&sms->pending,
 					__ofono_error_failed(sms->pending));
 		return;
@@ -263,6 +264,27 @@ static DBusMessage *sms_set_property(DBusConnection *conn, DBusMessage *msg,
 		return NULL;
 	}
 
+	if (!strcmp(property, "UseDeliveryReports")) {
+		const char *path = __ofono_atom_get_path(sms->atom);
+		dbus_bool_t value;
+
+		if (dbus_message_iter_get_arg_type(&var) != DBUS_TYPE_BOOLEAN)
+			return __ofono_error_invalid_args(msg);
+
+		dbus_message_iter_get_basic(&var, &value);
+
+		sms->use_delivery_reports = value;
+
+		g_dbus_send_reply(conn, msg, DBUS_TYPE_INVALID);
+
+		ofono_dbus_signal_property_changed(conn, path,
+						OFONO_SMS_MANAGER_INTERFACE,
+						"UseDeliveryReports",
+						DBUS_TYPE_BOOLEAN, &value);
+
+		return NULL;
+	}
+
 	return __ofono_error_invalid_args(msg);
 }
 
@@ -272,20 +294,20 @@ static void tx_finished(const struct ofono_error *error, int mr, void *data)
 	struct ofono_modem *modem = __ofono_atom_get_modem(sms->atom);
 	struct tx_queue_entry *entry = g_queue_peek_head(sms->txq);
 
-	ofono_debug("tx_finished");
+	DBG("tx_finished");
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
 		entry->retry += 1;
 
 		if (entry->retry != TXQ_MAX_RETRIES) {
-			ofono_debug("Sending failed, retry in %d secs",
+			DBG("Sending failed, retry in %d secs",
 					entry->retry * 5);
 			sms->tx_source = g_timeout_add_seconds(entry->retry * 5,
 								tx_next, sms);
 			return;
 		}
 
-		ofono_debug("Max retries reached, giving up");
+		DBG("Max retries reached, giving up");
 
 		entry = g_queue_pop_head(sms->txq);
 		__ofono_dbus_pending_reply(&entry->msg,
@@ -299,7 +321,7 @@ static void tx_finished(const struct ofono_error *error, int mr, void *data)
 		g_free(entry);
 
 		if (g_queue_peek_head(sms->txq)) {
-			ofono_debug("Previous send failed, scheduling next");
+			DBG("Previous send failed, scheduling next");
 			sms->tx_source = g_timeout_add(0, tx_next, sms);
 		}
 
@@ -325,7 +347,7 @@ static void tx_finished(const struct ofono_error *error, int mr, void *data)
 	g_free(entry);
 
 	if (g_queue_peek_head(sms->txq)) {
-		ofono_debug("Scheduling next");
+		DBG("Scheduling next");
 		sms->tx_source = g_timeout_add(0, tx_next, sms);
 	}
 }
@@ -333,7 +355,6 @@ static void tx_finished(const struct ofono_error *error, int mr, void *data)
 static gboolean tx_next(gpointer user_data)
 {
 	struct ofono_sms *sms = user_data;
-	time_t ts;
 	int send_mms = 0;
 	struct tx_queue_entry *entry = g_queue_peek_head(sms->txq);
 	struct pending_pdu *pdu = &entry->pdus[entry->cur_pdu];
@@ -341,21 +362,19 @@ static gboolean tx_next(gpointer user_data)
 
 	error.type = OFONO_ERROR_TYPE_NO_ERROR;
 
-	ofono_debug("tx_next: %p", entry);
+	DBG("tx_next: %p", entry);
 
 	sms->tx_source = 0;
 
 	if (!entry)
 		return FALSE;
 
-	ts = time(NULL);
-
-	if ((g_queue_get_length(sms->txq) > 1) &&
-			((ts - sms->last_mms) > 60))
+	if (g_queue_get_length(sms->txq) > 1
+			|| (entry->num_pdus - entry->cur_pdu) > 1)
 		send_mms = 1;
 
-	sms->driver->submit(sms, pdu->pdu, pdu->pdu_len, pdu->tpdu_len, send_mms,
-				tx_finished, sms);
+	sms->driver->submit(sms, pdu->pdu, pdu->pdu_len, pdu->tpdu_len,
+				send_mms, tx_finished, sms);
 
 	return FALSE;
 }
@@ -393,13 +412,26 @@ static struct tx_queue_entry *create_tx_queue_entry(GSList *msg_list)
 
 		sms_encode(s, &pdu->pdu_len, &pdu->tpdu_len, pdu->pdu);
 
-		ofono_debug("pdu_len: %d, tpdu_len: %d",
+		DBG("pdu_len: %d, tpdu_len: %d",
 				pdu->pdu_len, pdu->tpdu_len);
 	}
 
 	return entry;
 }
 
+/*
+ * Pre-process a SMS text message and deliver it [D-Bus SendMessage()]
+ *
+ * @conn: D-Bus connection
+ * @msg: message data (telephone number and text)
+ * @data: SMS object to use for transmision
+ *
+ * An alphabet is chosen for the text and it (might be) segmented in
+ * fragments by sms_text_prepare() into @msg_list. A queue list @entry
+ * is created by create_tx_queue_entry() and g_queue_push_tail()
+ * appends that entry to the SMS transmit queue. Then the tx_next()
+ * function is scheduled to run to process the queue.
+ */
 static DBusMessage *sms_send_message(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
@@ -419,12 +451,13 @@ static DBusMessage *sms_send_message(DBusConnection *conn, DBusMessage *msg,
 	if (valid_phone_number_format(to) == FALSE)
 		return __ofono_error_invalid_format(msg);
 
-	msg_list = sms_text_prepare(text, 0, TRUE, &ref_offset);
+	msg_list = sms_text_prepare(text, 0, TRUE, &ref_offset,
+					sms->use_delivery_reports);
 
 	if (!msg_list)
 		return __ofono_error_invalid_format(msg);
 
-	ofono_debug("ref: %d, offset: %d", sms->ref, ref_offset);
+	DBG("ref: %d, offset: %d", sms->ref, ref_offset);
 
 	set_ref_and_to(msg_list, sms->ref, ref_offset, to);
 	entry = create_tx_queue_entry(msg_list);
@@ -474,9 +507,9 @@ static GDBusSignalTable sms_manager_signals[] = {
 static void dispatch_app_datagram(struct ofono_sms *sms, int dst, int src,
 					unsigned char *buf, long len)
 {
-	ofono_debug("Got app datagram for dst port: %d, src port: %d",
+	DBG("Got app datagram for dst port: %d, src port: %d",
 			dst, src);
-	ofono_debug("Contents-Len: %ld", len);
+	DBG("Contents-Len: %ld", len);
 }
 
 static void dispatch_text_message(struct ofono_sms *sms,
@@ -506,7 +539,7 @@ static void dispatch_text_message(struct ofono_sms *sms,
 	else
 		signal_name = "IncomingMessage";
 
-	signal = dbus_message_new_signal(path, SMS_MANAGER_INTERFACE,
+	signal = dbus_message_new_signal(path, OFONO_SMS_MANAGER_INTERFACE,
 						signal_name);
 
 	if (!signal)
@@ -714,7 +747,7 @@ void ofono_sms_deliver_notify(struct ofono_sms *sms, unsigned char *pdu,
 	}
 
 	if (s.deliver.pid == SMS_PID_TYPE_SM_TYPE_0) {
-		ofono_debug("Explicitly ignoring type 0 SMS");
+		DBG("Explicitly ignoring type 0 SMS");
 		return;
 	}
 
@@ -845,8 +878,8 @@ static void sms_unregister(struct ofono_atom *atom)
 	struct ofono_modem *modem = __ofono_atom_get_modem(atom);
 	const char *path = __ofono_atom_get_path(atom);
 
-	g_dbus_unregister_interface(conn, path, SMS_MANAGER_INTERFACE);
-	ofono_modem_remove_interface(modem, SMS_MANAGER_INTERFACE);
+	g_dbus_unregister_interface(conn, path, OFONO_SMS_MANAGER_INTERFACE);
+	ofono_modem_remove_interface(modem, OFONO_SMS_MANAGER_INTERFACE);
 
 	if (sms->mw_watch) {
 		__ofono_modem_remove_atom_watch(modem, sms->mw_watch);
@@ -888,6 +921,9 @@ static void sms_remove(struct ofono_atom *atom)
 					"NextMessageId", sms->next_msg_id);
 		g_key_file_set_integer(sms->settings, SETTINGS_GROUP,
 					"NextReference", sms->ref);
+		g_key_file_set_boolean(sms->settings, SETTINGS_GROUP,
+					"UseDeliveryReports",
+					sms->use_delivery_reports);
 
 		storage_close(sms->imsi, SETTINGS_STORE, sms->settings, TRUE);
 
@@ -899,6 +935,17 @@ static void sms_remove(struct ofono_atom *atom)
 	g_free(sms);
 }
 
+
+/*
+ * Create a SMS driver
+ *
+ * This creates a SMS driver that is hung off a @modem
+ * object. However, for the driver to be used by the system, it has to
+ * be registered with the oFono core using ofono_sms_register().
+ *
+ * This is done once the modem driver determines that SMS is properly
+ * supported by the hardware.
+ */
 struct ofono_sms *ofono_sms_create(struct ofono_modem *modem,
 					unsigned int vendor,
 					const char *driver,
@@ -963,12 +1010,22 @@ static void sms_load_settings(struct ofono_sms *sms, const char *imsi)
 							"NextMessageId", NULL);
 	sms->ref = g_key_file_get_integer(sms->settings, SETTINGS_GROUP,
 							"NextReference", NULL);
+	sms->use_delivery_reports =
+		g_key_file_get_boolean(sms->settings, SETTINGS_GROUP,
+					"UseDeliveryReports", NULL);
 
 	if (sms->ref >= 65536)
 		sms->ref = 1;
-
 }
 
+
+/*
+ * Indicate oFono that a SMS driver is ready for operation
+ *
+ * This is called after ofono_sms_create() was done and the modem
+ * driver determined that a modem supports SMS correctly. Once this
+ * call succeeds, the D-BUS interface for SMS goes live.
+ */
 void ofono_sms_register(struct ofono_sms *sms)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
@@ -978,16 +1035,16 @@ void ofono_sms_register(struct ofono_sms *sms)
 	struct ofono_atom *sim_atom;
 
 	if (!g_dbus_register_interface(conn, path,
-					SMS_MANAGER_INTERFACE,
+					OFONO_SMS_MANAGER_INTERFACE,
 					sms_manager_methods,
 					sms_manager_signals,
 					NULL, sms, NULL)) {
 		ofono_error("Could not create %s interface",
-				SMS_MANAGER_INTERFACE);
+				OFONO_SMS_MANAGER_INTERFACE);
 		return;
 	}
 
-	ofono_modem_add_interface(modem, SMS_MANAGER_INTERFACE);
+	ofono_modem_add_interface(modem, OFONO_SMS_MANAGER_INTERFACE);
 
 	sms->mw_watch = __ofono_modem_add_atom_watch(modem,
 					OFONO_ATOM_TYPE_MESSAGE_WAITING,
